/*
 * Copyright (C) 2012 Giorgio Vazzana
 *
 * This file is part of parole-conference.
 *
 * parole-conference is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * parole-conference is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "network.h"
#include "nc.h"
#include "rw.h"

#define DEBUG 0

static int is_same_addr(struct sockaddr_in a, struct sockaddr_in b)
{
	if (a.sin_addr.s_addr == b.sin_addr.s_addr && a.sin_port == b.sin_port)
		return 1;
	else
		return 0;
}


int resolve_address(struct sockaddr_in *addr, char *host, int port)
{
	struct hostent *hostinfo = NULL;
	char **ppchar;

#define PRINT snprintf(msgbuf, MBS,

	/* let's see if we already have a valid ip adress */
	if (inet_aton(host, &addr->sin_addr)) {
		PRINT "Host: %s\n", host);
		print_msg(msgbuf);
	} else {
		/* get host info */
		hostinfo = gethostbyname(host);
		if (!hostinfo) {
			PRINT "Cannot get info for host: %s\n", host);
			print_msg(msgbuf);
			return -1;
		}

		/* print host info */
		PRINT "Host: %s\n", host);
		print_msg(msgbuf);
		PRINT "  Name: %s\n", hostinfo->h_name);
		print_msg(msgbuf);
/*		PRINT "  Aliases:");
		print_msg(msgbuf);
		for (ppchar = hostinfo->h_aliases; *ppchar; ppchar++) {
			PRINT " %s", *ppchar);
			print_msg(msgbuf);
		}
		PRINT "\n");
		print_msg(msgbuf);
		PRINT "  Type: %s\n", hostinfo->h_addrtype == AF_INET ? "AF_INET" : "AF_INET6");
		print_msg(msgbuf);*/
		PRINT "  Address list:");
		print_msg(msgbuf);
		for (ppchar = hostinfo->h_addr_list; *ppchar; ppchar++) {
			PRINT " %s", inet_ntoa(*(struct in_addr *)*ppchar) );
			print_msg(msgbuf);
		}
		PRINT "\n");
		print_msg(msgbuf);

		if (hostinfo->h_addrtype == AF_INET6) {
			PRINT "AF_INET6 not supported.\n");
			print_msg(msgbuf);
			return -1;
		}
	}

	PRINT "Port: %d\n\n", port);
	print_msg(msgbuf);

	/* fill in address */
	addr->sin_family   = AF_INET;
	if (hostinfo)
		addr->sin_addr = *(struct in_addr *) *hostinfo->h_addr_list;
	addr->sin_port     = htons(port);

	return 0;
#undef PRINT
}

void send_udp_packet(int sockfd, struct udp_packet_st *udp_packet)
{
	ssize_t nw;

	nw = sendto(sockfd, udp_packet->data, udp_packet->len, 0,
		        (struct sockaddr *)&udp_packet->addr, sizeof(udp_packet->addr));
	if (nw == -1) {
		snprintf(msgbuf, MBS, "sendto(): %s\n", strerror(errno));
		die(msgbuf, 1);
	}
	if (nw < udp_packet->len) {
		snprintf(msgbuf, MBS, "sendto() returned %zd, expected %zd\n", nw, udp_packet->len);
		log_msg(msgbuf);
	}
}

void receive_udp_packet(int sockfd, struct udp_packet_st *udp_packet)
{
	ssize_t   nr;
	socklen_t addr_len;

	udp_packet->len = UDP_PACKET_MAXLEN;
	addr_len = sizeof(udp_packet->addr);
	nr = recvfrom(sockfd, udp_packet->data, udp_packet->len, 0,
		          (struct sockaddr *)&udp_packet->addr, &addr_len);
	if (nr == -1) {
		snprintf(msgbuf, MBS, "recvfrom(): %s\n", strerror(errno));
		die(msgbuf, 1);
	}

	udp_packet->len = nr;
}


int node_get_count(struct node *nodes[])
{
	int i, n;

	n = 0;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i])
			n++;
	}
	return n;
}

int node_search(struct node *nodes[], struct sockaddr_in addr)
{
	int i, idx;

	idx = -1;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] && is_same_addr(addr, nodes[i]->addr)) {
			idx = i;
			break;
		}
	}
	return idx;
}

void node_delete(struct node *nodes[], int idx, int for_timeout, struct tone *tone)
{
	snprintf(msgbuf, MBS, "Node %s:%hu (%s) left%s\n",
	        inet_ntoa(nodes[idx]->addr.sin_addr), ntohs(nodes[idx]->addr.sin_port),
		    nodes[idx]->nick, for_timeout ? " (timeout)" : " (call ended)");
	print_msg(msgbuf);
	tone->tone_play = (tone->tone_play > 10*ONE_OPUS_FRAME) ? tone->tone_play : 10*ONE_OPUS_FRAME;

	/* destroy decoder state */
	opus_decoder_destroy(nodes[idx]->dec);

	free(nodes[idx]);
	nodes[idx] = NULL;
}

int node_add(struct node *nodes[], struct sockaddr_in addr, const char *nick, struct tone *tone)
{
	int i, idx, error;

	/* check if we have space for another node */
	idx = -1;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] == NULL) {
			idx = i;
			break;
		}
	}
	if (idx == -1) {
		snprintf(msgbuf, MBS, "Error, max node capacity reached (%d nodes).\n", MAX_NUMBER_OF_NODES);
		print_msg(msgbuf);
		return -1;
	}

	/* allocate and clear memory for new node */
	nodes[idx] = calloc(1, sizeof(struct node));
	if (nodes[idx] == NULL) {
		snprintf(msgbuf, MBS, "Error: calloc() failed.\n");
		die(msgbuf, 1);
	}

	/* set non-zero fields */
	nodes[idx]->addr                 = addr;
	if (nick != NULL)
		memcpy(nodes[idx]->nick, nick, NICKLEN);
	else
		memset(nodes[idx]->nick,    0, NICKLEN);
	nodes[idx]->timeout              = TIMEOUT;

	/* create decoder state */
	nodes[idx]->dec = opus_decoder_create(rate, channels, &error);
	if (error != OPUS_OK) {
		snprintf(msgbuf, MBS, "opus error: %s\n", opus_strerror(error));
		die(msgbuf, 1);
	}

	snprintf(msgbuf, MBS, "Node %s:%hu (%s) joined the network\n",
	        inet_ntoa(nodes[idx]->addr.sin_addr), ntohs(nodes[idx]->addr.sin_port),
	        nodes[idx]->nick);
	print_msg(msgbuf);
	tone->tone_play = (tone->tone_play > 20*ONE_OPUS_FRAME) ? tone->tone_play : 20*ONE_OPUS_FRAME;

	return idx;
}


void packet_call(struct udp_packet_st *udp_packet, char *nick)
{
	udp_packet->len = 2+NICKLEN;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_CALL);
	memcpy(udp_packet->data+2, nick, NICKLEN);
}

void packet_refuse(struct udp_packet_st *udp_packet)
{
	udp_packet->len = 2;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_REFUSE);
}

void packet_table(struct udp_packet_st *udp_packet, char *nick, struct node *nodes[], int exclude_idx)
{
	int i;
	uint8_t *d;
	size_t   l;
	uint16_t table_size;

	d = udp_packet->data;
	l = 0;

	/* all nodes excluding the sender of the call request */
	table_size = node_get_count(nodes);

	/* nodes address table */
	write_be16(d, UDP_PACKET_TYPE_TABLE); d+=2; l+=2;
	/* table size */
	write_be16(d, table_size);            d+=2; l+=2;
	/* my nick */
	memcpy(d, nick, NICKLEN);             d+=NICKLEN; l+=NICKLEN;
	/* other nodes except packet sender */
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] == NULL)
			continue;
		if (i == exclude_idx)
			continue;
		memcpy(d, nodes[i]->nick, NICKLEN);                   d+=NICKLEN; l+=NICKLEN;
		write_be32(d, ntohl(nodes[i]->addr.sin_addr.s_addr)); d+=4; l+=4;
		write_be16(d, ntohs(nodes[i]->addr.sin_port));        d+=2; l+=2;
	}
	udp_packet->len = l;
}

void packet_audio(struct udp_packet_st *udp_packet, uint32_t sequence_number, uint8_t *payload, size_t payloadlen)
{
	udp_packet->len = UDP_PACKET_AUDIO_HEADER_LEN + payloadlen;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_AUDIO);
	write_be32(udp_packet->data+2, sequence_number);
	memcpy(udp_packet->data + UDP_PACKET_AUDIO_HEADER_LEN, payload, payloadlen);
}

void packet_chat(struct udp_packet_st *udp_packet, char *buf, size_t len)
{
	udp_packet->len = 2+len;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_CHAT);
	memcpy(udp_packet->data+2, buf, len);
}

void packet_nop(struct udp_packet_st *udp_packet)
{
	udp_packet->len = 2;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_NOP);
}

void packet_end(struct udp_packet_st *udp_packet)
{
	udp_packet->len = 2;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_END);
}


void send_packet_to_all(struct node *nodes[], int sockfd, struct udp_packet_st *udp_packet, size_t *bytes_sent)
{
	int i;

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] && nodes[i]->knows_about_me) {
			udp_packet->addr = nodes[i]->addr;
			send_udp_packet(sockfd, udp_packet);
			*bytes_sent += udp_packet->len;
		}
	}
}

void call_new_nodes(struct node *nodes[], int sockfd, char *nick, int use_relay)
{
	int i, j, relay_count;
	uint8_t *d;
	size_t   l;
	struct udp_packet_st udp_packet_response;

	/* send call requests to those who don't know about me */
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] && nodes[i]->knows_about_me == 0) {
			udp_packet_response.addr = nodes[i]->addr;
			packet_call(&udp_packet_response, nick);
			send_udp_packet(sockfd, &udp_packet_response);

			if (use_relay) {
				/* prepare packet for relay */
				d = udp_packet_response.data;
				l = 0;
				memmove(d+8, d, udp_packet_response.len);

				write_be16(d, UDP_PACKET_TYPE_RELAY);                 d+=2; l+=2;
				write_be32(d, ntohl(nodes[i]->addr.sin_addr.s_addr)); d+=4; l+=4;
				write_be16(d, ntohs(nodes[i]->addr.sin_port));        d+=2; l+=2;
				udp_packet_response.len += l;

				/* send this packet to (at most) 2 other nodes that can relay it */
				relay_count = 2;
				for (j = 0; j < MAX_NUMBER_OF_NODES; j++) {
					if (nodes[j] && nodes[j]->knows_about_me && j != i) {
						udp_packet_response.addr = nodes[j]->addr;
						send_udp_packet(sockfd, &udp_packet_response);
						relay_count--;
						if (relay_count == 0)
							break;
					}
				}

				if (relay_count < 2) {
					snprintf(msgbuf, MBS, "I was not able to contact %s:%hu (%s), using relay\n",
							inet_ntoa(nodes[i]->addr.sin_addr), ntohs(nodes[i]->addr.sin_port),
							nodes[i]->nick);
					print_msg(msgbuf);
				}
			}
		}
	}
}

struct address_book {
	int                n;
	char               nick[MAX_NUMBER_OF_NODES][NICKLEN];
	struct sockaddr_in addr[MAX_NUMBER_OF_NODES];
};

int process_packet(struct PContext *pctx, struct udp_packet_st *udp_packet)
{
	int i, j, error, idx;
	uint8_t *d;
	size_t   l;
	struct node **nodes;
	struct tone  *tone;
	uint16_t packet_type, table_size;
	struct address_book ab;
	struct udp_packet_st udp_packet_response;

	uint8_t    *opus_dec_packet;
	opus_int32  opus_dec_packetlen;
	opus_int32  pcmlen;

	nodes = pctx->nodes;
	tone  = &pctx->tone;
	d = udp_packet->data;
	l = udp_packet->len;

	if (udp_packet->len < 2)
		return RES_ERROR;

	packet_type = read_be16(d); d+=2; l-=2;

	/* do we know this node already? */
	idx = node_search(nodes, udp_packet->addr);
	/* reset timeout counter */
	if (idx != -1)
		nodes[idx]->timeout = TIMEOUT;

	switch (packet_type) {
	case UDP_PACKET_TYPE_CALL:
		if (l != NICKLEN) {
			snprintf(msgbuf, MBS, "Invalid CALL packet received\n");
			print_msg(msgbuf);
			return RES_ERROR;
		}
/*
		snprintf(msgbuf, MBS, "Call from %s:%hu (%s)\n",
		         inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port), d);
		print_msg(msgbuf);
*/
#if 0
		/* ask if we can accept the call */
		if (pctx->autoaccept == 0) {
			if (pctx->answer == ANS_NOTPRESENT) {
				snprintf(msgbuf, MBS, "Press [F6] to accept, [F7] to refuse.\n");
				print_msg(msgbuf);
				tone->tone_play = TONEBUFLEN;
				pctx->saved_packet   = *udp_packet;
				pctx->need_answer    = 1;
				return RES_NEED_ANSWER;
			} else {
				if (pctx->answer != ANS_YES) {
					udp_packet_response.addr = udp_packet->addr;
					packet_refuse(&udp_packet_response);
					send_udp_packet(pctx->sockfd, &udp_packet_response);
					pctx->need_answer = 0;
					pctx->answer      = ANS_NOTPRESENT;
					snprintf(msgbuf, MBS, "Call refused.\n");
					print_msg(msgbuf);
					return RES_REFUSED;
				} else {
					pctx->need_answer = 0;
					pctx->answer      = ANS_NOTPRESENT;
					snprintf(msgbuf, MBS, "Call accepted.\n");
					print_msg(msgbuf);
				}
			}
		}
#endif
		/* insert remote node */
		if (idx == -1) {
			idx = node_add(nodes, udp_packet->addr, (char *)d, tone);
			if (idx < 0) {
				udp_packet_response.addr = udp_packet->addr;
				packet_refuse(&udp_packet_response);
				send_udp_packet(pctx->sockfd, &udp_packet_response);
				return RES_ERROR;
			}
			nodes[idx]->knows_about_me = 1;
		}

		/* send table */
		udp_packet_response.addr = udp_packet->addr;
		packet_table(&udp_packet_response, pctx->nick, nodes, idx);
		send_udp_packet(pctx->sockfd, &udp_packet_response);
		return RES_OK;


	case UDP_PACKET_TYPE_REFUSE:
		if (idx == -1) {
			snprintf(msgbuf, MBS, "Unexpected REFUSE packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			print_msg(msgbuf);
			return RES_ERROR;
		}
		snprintf(msgbuf, MBS, "Node %s:%hu (%s) refused the call\n",
			    inet_ntoa(nodes[idx]->addr.sin_addr), ntohs(nodes[idx]->addr.sin_port),
			    nodes[idx]->nick);
		print_msg(msgbuf);
		node_delete(nodes, idx, 0, tone);
		return RES_REFUSED;


	case UDP_PACKET_TYPE_TABLE:
		if (idx == -1) {
			snprintf(msgbuf, MBS, "Unexpected TABLE packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			print_msg(msgbuf);
			return RES_ERROR;
		}

		table_size = read_be16(d); d+=2; l-=2;
		if (l != (NICKLEN*table_size + 6*(table_size-1))) {
			snprintf(msgbuf, MBS, "Invalid TABLE packet received\n");
			print_msg(msgbuf);
			return RES_ERROR;
		}

		/* copy table to addres book */
		ab.n = table_size;
		for (i = 0; i < ab.n; i++) {
			memcpy(ab.nick[i], d, NICKLEN);                       d+=NICKLEN; l-=NICKLEN;
			if (i == 0) {
				ab.addr[i] = udp_packet->addr;
			} else {
				ab.addr[i].sin_family      = AF_INET;
				ab.addr[i].sin_addr.s_addr = htonl(read_be32(d)); d+=4; l-=4;
				ab.addr[i].sin_port        = htons(read_be16(d)); d+=2; l-=2;
			}
		}

		/* copy sender's nick */
		memcpy(nodes[idx]->nick, ab.nick[0], NICKLEN);

		/* packet sender knows about me */
		nodes[idx]->knows_about_me = 1;

		snprintf(msgbuf, MBS, "Node %s:%hu (%s) accepted the call\n",
			    inet_ntoa(nodes[idx]->addr.sin_addr), ntohs(nodes[idx]->addr.sin_port),
			    nodes[idx]->nick);
		print_msg(msgbuf);

		/* insert unknown nodes */
		for (j = 0; j < ab.n; j++) {
			int tidx;

			tidx = node_search(nodes, ab.addr[j]);
			if (tidx == -1)
				node_add(nodes, ab.addr[j], ab.nick[j], tone);
		}

		/* send call requests to newly added nodes. we don't use relay initially */
		call_new_nodes(nodes, pctx->sockfd, pctx->nick, 0);
		return RES_OK;


	case UDP_PACKET_TYPE_AUDIO:
		if (idx == -1) {
#if DEBUG
			snprintf(msgbuf, MBS, "Unexpected AUDIO packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			print_msg(msgbuf);
#endif
			return RES_ERROR;
		}

		/* read packet header */
		nodes[idx]->sequence_number = read_be32(d);
		d+=4; l-=4;

		/* ok, sequence numbers of packets must always increase */
		if (nodes[idx]->sequence_number > nodes[idx]->last_sequence_number ||
		    nodes[idx]->packets_count == 0) {
			/* check if we lost packets */
			if (nodes[idx]->sequence_number != nodes[idx]->last_sequence_number + 1 &&
			    nodes[idx]->packets_count != 0) {
				error = nodes[idx]->sequence_number - (nodes[idx]->last_sequence_number + 1);
#if DEBUG
				snprintf(msgbuf, MBS, "Oops, %d audio packets lost.\n", error);
				print_msg(msgbuf);
#endif
				nodes[idx]->packets_lost += error;
			}

			nodes[idx]->last_sequence_number = nodes[idx]->sequence_number;

			opus_dec_packet    = udp_packet->data + UDP_PACKET_AUDIO_HEADER_LEN;
			opus_dec_packetlen = udp_packet->len  - UDP_PACKET_AUDIO_HEADER_LEN;

			/* decode */
			if (PCMBUFLEN - nodes[idx]->pcmlen < ONE_OPUS_FRAME) {
#if DEBUG
				snprintf(msgbuf, MBS, "Warning: packet from '%s', pcmbuf is full (%d)\n",
				        nodes[idx]->nick, nodes[idx]->pcmlen);
				print_msg(msgbuf);
#endif
				nodes[idx]->pcmfull++;
				nodes[idx]->pcmlen -= ONE_OPUS_FRAME;
				memmove(nodes[idx]->pcm, nodes[idx]->pcm + channels * ONE_OPUS_FRAME,
				        nodes[idx]->pcmlen * channels * sizeof(nodes[idx]->pcm[0]));
			}

			pcmlen = opus_decode(nodes[idx]->dec, opus_dec_packet, opus_dec_packetlen,
			                     nodes[idx]->pcm + channels * nodes[idx]->pcmlen,
			                     PCMBUFLEN - nodes[idx]->pcmlen, 0);
			if (pcmlen < 0) {
				snprintf(msgbuf, MBS, "opus_decode() failed: %s\n", opus_strerror(pcmlen));
				die(msgbuf, 1);
			}
			if (pcmlen == ONE_OPUS_FRAME) {
				nodes[idx]->pcmlen += pcmlen;
			} else {
				snprintf(msgbuf, MBS, "Warning: pcmlen = %d, discarding pcm data\n", pcmlen);
				print_msg(msgbuf);
			}
			/* increment packets counter */
			nodes[idx]->packets_count++;
		} else {
#if DEBUG
			snprintf(msgbuf, MBS, "Oops, audio packet duplicated or out of order.\n");
			print_msg(msgbuf);
#endif
		}
		return RES_OK;


	case UDP_PACKET_TYPE_CHAT:
		if (idx == -1) {
			snprintf(msgbuf, MBS, "Unexpected CHAT packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			print_msg(msgbuf);
			return RES_ERROR;
		}

		snprintf(msgbuf, MBS, "%s> %s", nodes[idx]->nick, d);
		wattron(nctx.win, A_BOLD);
		print_msg(msgbuf);
		wattroff(nctx.win, A_BOLD);
		tone->tone_play = (tone->tone_play > 5*ONE_OPUS_FRAME) ? tone->tone_play : 5*ONE_OPUS_FRAME;
		return RES_OK;


	case UDP_PACKET_TYPE_NOP:
		return RES_OK;


	case UDP_PACKET_TYPE_RELAY:
		if (l < 6) {
			snprintf(msgbuf, MBS, "Invalid RELAY packet received\n");
			print_msg(msgbuf);
			return RES_ERROR;
		}

		/* copy received packet */
		memcpy(udp_packet_response.data, udp_packet->data, udp_packet->len);
		udp_packet_response.len = udp_packet->len;

		/* extract destination */
		udp_packet_response.addr.sin_family      = AF_INET;
		udp_packet_response.addr.sin_addr.s_addr = htonl(read_be32(d)); d+=4; l-=4;
		udp_packet_response.addr.sin_port        = htons(read_be16(d)); d+=2; l-=2;

		/* this now becomes a relayed packet, save source address */
		d = udp_packet_response.data;
		l = 0;
		write_be16(d, UDP_PACKET_TYPE_RELAYED);                 d+=2; l+=2;
		write_be32(d, ntohl(udp_packet->addr.sin_addr.s_addr)); d+=4; l+=4;
		write_be16(d, ntohs(udp_packet->addr.sin_port));        d+=2; l+=2;

		send_udp_packet(pctx->sockfd, &udp_packet_response);
		return RES_OK;


	case UDP_PACKET_TYPE_RELAYED:
		if (l < 6) {
			snprintf(msgbuf, MBS, "Invalid RELAYED packet received\n");
			print_msg(msgbuf);
			return RES_ERROR;
		}

		/* copy received packet, discarding relayed header */
		memcpy(udp_packet_response.data, udp_packet->data+8, udp_packet->len-8);
		udp_packet_response.len = udp_packet->len-8;

		/* extract original source */
		udp_packet_response.addr.sin_family      = AF_INET;
		udp_packet_response.addr.sin_addr.s_addr = htonl(read_be32(d)); d+=4; l-=4;
		udp_packet_response.addr.sin_port        = htons(read_be16(d)); d+=2; l-=2;

		return process_packet(pctx, &udp_packet_response);


	case UDP_PACKET_TYPE_END:
		if (idx == -1) {
			snprintf(msgbuf, MBS, "Unexpected END packet from: %s:%hu\n",
			        inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			print_msg(msgbuf);
			return RES_ERROR;
		}
		node_delete(nodes, idx, 0, tone);
		return RES_OK;


	default:
		snprintf(msgbuf, MBS, "Unknown packet type\n");
		print_msg(msgbuf);
		break;
	}
	return RES_ERROR;
}
