/*
 * Copyright (C) 2012 Giorgio Vazzana
 *
 * This file is part of parole-conference.
 *
 * parole-conference is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * parole-conference is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <poll.h>
#include <time.h>
#include <locale.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "parole-conference.h"
#include "nc.h"
#include "audio.h"
#include "network.h"

#define VERSION "0.10alpha10"

const unsigned int rate = 48000, channels = 2;
char               msgbuf[MBS];
FILE              *fplog;
struct NContext    nctx;

static volatile sig_atomic_t keep_going;
static volatile sig_atomic_t window_resized;

static void signal_handler(int sig)
{
	if (sig == SIGINT)
		keep_going = 0;
	if (sig == SIGWINCH)
		window_resized = 1;
}

static void check_opus_error_or_die(int error)
{
	if (error != OPUS_OK) {
		snprintf(msgbuf, MBS, "opus error: %s\n", opus_strerror(error));
		die(msgbuf, 1);
	}
}

static opus_int32 convert_opus_bw(opus_int32 bandwidth)
{
	switch (bandwidth) {
	case OPUS_BANDWIDTH_NARROWBAND:
		return 4000;
	case OPUS_BANDWIDTH_MEDIUMBAND:
		return 6000;
	case OPUS_BANDWIDTH_WIDEBAND:
		return 8000;
	case OPUS_BANDWIDTH_SUPERWIDEBAND:
		return 12000;
	case OPUS_BANDWIDTH_FULLBAND:
		return 20000;
	default: /* AUTO */
		return 0;
	}
}

static void update_node_stats(struct PContext *pctx)
{
	int i;

	pctx->packets_lost_g  = 0;
	pctx->packets_count_g = 0;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i]) {
			pctx->packets_lost_g  += pctx->nodes[i]->packets_lost;
			pctx->packets_count_g += pctx->nodes[i]->packets_count;
		}
	}
}

void generate_tone(struct tone *tone)
{
	int i;
	double v;

	for (i = 0; i < TONEBUFLEN; i++) {
		v = 10000.0 * sin(2.0 * M_PI * 440.0 * i / (double)rate);
		if (channels == 1) {
			tone->pcm[i] = lrint(v);
		} else {
			tone->pcm[channels*i    ] = lrint(v);
			tone->pcm[channels*i + 1] = lrint(v);
		}
	}
	tone->tone_pos  = 0;
	tone->tone_play = 0;
}

static int32_t find_peak_int32(int32_t *pcm, size_t len)
{
	size_t i;
	int32_t abs_left, abs_right;
	int32_t peak_left = 0, peak_right = 0;

	for (i = 0; i < len; i+=2) {
		abs_left  = abs(pcm[i  ]);
		abs_right = abs(pcm[i+1]);
		if (abs_left > peak_left)
			peak_left = abs_left;
		if (abs_right > peak_right)
			peak_right = abs_right;
	}
	return (peak_left > peak_right) ? peak_left : peak_right;
}

static int mix_audio(struct PContext *pctx)
{
	int i, j, n;
	opus_int32 pcm32[2*ONE_OPUS_FRAME], peak;
	float gain;

	memset(pcm32, 0, sizeof(pcm32));

	n = 0;
	if (pctx->tone.tone_pos >= pctx->tone.tone_play) {
		pctx->tone.tone_pos  = 0;
		pctx->tone.tone_play = 0;
	} else {
		n++;

		/* sum (mix) pcm from tone buffer */
		for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
			pcm32[j] += pctx->tone.pcm[channels * pctx->tone.tone_pos + j];

		pctx->tone.tone_pos += ONE_OPUS_FRAME;
	}

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i] && pctx->nodes[i]->pcmlen >= ONE_OPUS_FRAME) {
			n++;

			/* sum (mix) pcm from different nodes */
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pcm32[j] += pctx->nodes[i]->pcm[j];

			/* remove one opus frame from the head of the queue */
			pctx->nodes[i]->pcmlen -= ONE_OPUS_FRAME;
			if (pctx->nodes[i]->pcmlen)
				memmove(pctx->nodes[i]->pcm, pctx->nodes[i]->pcm + channels * ONE_OPUS_FRAME,
				        pctx->nodes[i]->pcmlen * channels * sizeof(pctx->nodes[i]->pcm[0]));
		}
	}

	if (n) {
		peak = find_peak_int32(pcm32, channels * ONE_OPUS_FRAME);

		gain = (float)(INT16_MAX) / (float)peak;
		gain = (gain > 1.0f) ? 1.0f : gain;

		if (gain != 1.0f) {
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pctx->pcm[j] = lrintf((float)pcm32[j] * gain);
		} else {
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pctx->pcm[j] = pcm32[j];
		}

		pctx->pcmlen = ONE_OPUS_FRAME;
	} else {
		pctx->pcmlen = 0;
	}
	return n;
}

static void usage()
{
	fprintf(stderr,
"Usage: parole-conference [options]\n"
"Options:\n"
"       -h             print help\n"
"       -v             increment verbosity level\n"
"       -n nick        set nickname\n"
"       -p port        listening port            (8110)\n"
"       -b bitrate     set bitrate in bit/s      (16000)\n"
"       -d audiodev    use audio device audiodev (hw:0,0)\n"
/*"       -a             autoaccept calls\n"*/
	);
}

int main(int argc, char *argv[])
{
	int                  ret, opt, i, error;
	int                  verbose = 0;
	struct sigaction     act;
	struct pollfd        pfds[6];
	int                  nfds, pollret;
	int                  ic_total, ic_talk, ic_mute;

	/* context */
	struct PContext      pctx;

	/* udp socket */
	int                  local_port = 8110;
	struct sockaddr_in   local_addr;
	struct udp_packet_st udp_packet;
	size_t               total_bytes_out;
	size_t               total_bytes_in;

	/* audio data */
	struct audio_data    ad_capture, ad_playback;
	char                *audiodevice = "hw:0,0";
	float                micgain, peak_percent;
	int                  keep_going_capture, overruns, underruns, micmute;

	/* opus encoder */
	OpusEncoder         *enc;
	opus_int32           bitrate = 16000, bandwidth;
//	opus_int16           inbuffer[2*960];	/* 2ch, 20ms at 48kHz (FS * 0.02) */
	uint8_t              opus_enc_packet[4000];
	opus_int32           opus_enc_packetlen;


	setlocale(LC_ALL, "");
	srand(time(NULL));
	fprintf(stderr, "parole-conference, ver. %s\n\n", VERSION);

	/* clear parole context; in particular it'll clear nick, options and array of nodes */
	memset(&pctx, 0, sizeof(pctx));
	/* clear ncurses context */
	memset(&nctx, 0, sizeof(nctx));

	generate_tone(&pctx.tone);


	/* parse command line options */
	while ((opt = getopt(argc, argv, "hvn:p:b:d:a")) != -1) {
		switch (opt) {
		case 'h':
			usage();
			exit(0);
			break;
		case 'v':
			verbose++;
			break;
		case 'n':
			strncpy(pctx.nick, optarg, NICKLEN-1);
			pctx.nick[NICKLEN-1] = '\0';
			break;
		case 'p':
			local_port = atoi(optarg);
			break;
		case 'b':
			bitrate = atoi(optarg);
			break;
		case 'd':
			audiodevice = optarg;
			break;
/*		case 'a':
			pctx.autoaccept = 1;
			break;*/
		default:
			fprintf(stderr, "\n");
			usage();
			exit(1);
		}
	}

#if 0
	fprintf(stderr, "verbose=%d, nick=%s, local_port=%d, bitrate=%d, audiodevice=%s, autoaccept=%d\n"
	        verbose, pctx.nick, local_port, bitrate, audiodevice, pctx.autoaccept);
#endif

	fplog = fopen(LOGFILE, "w");
	if (fplog == NULL) {
		fprintf(stderr, "Error: cannot open log file '%s'.\n", LOGFILE);
		exit(1);
	}

	/* if no nick was specified, choose a random one */
	if (pctx.nick[0] == '\0') {
		memcpy(pctx.nick, "guest", 5);
		for (i = 0; i < 3; i++)
			pctx.nick[i+5] = '0' + (rand() % 10);
	}

	/* init ncurses */
	initscr();
	if (has_colors() == FALSE) {
		snprintf(msgbuf, MBS, "Your terminal does not support color\n");
		die(msgbuf, 1);
	}
	start_color();
	assume_default_colors(COLOR_WHITE, COLOR_BLUE);
	init_pair(1, COLOR_BLACK,  COLOR_WHITE);
	init_pair(2, COLOR_GREEN,  COLOR_WHITE);
	init_pair(3, COLOR_YELLOW, COLOR_WHITE);
	init_pair(4, COLOR_RED,    COLOR_WHITE);
	cbreak();
	noecho();
	keypad(stdscr, TRUE);
	getmaxyx(stdscr, nctx.rows, nctx.cols);

	/* create message window */
	nctx.win = newwin(nctx.rows - 4, nctx.cols, 2, 0);
	if (nctx.win == NULL) {
		snprintf(msgbuf, MBS, "Could not create ncurses window\n");
		die(msgbuf, 1);
	}
	scrollok(nctx.win, TRUE);

	print_bars(VERSION);
	print_prompt(pctx.nick);

	/* install signal handler for window resize */
	act.sa_handler = signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	ret = sigaction(SIGWINCH, &act, NULL);
	if (ret == -1) {
		snprintf(msgbuf, MBS, "Error: sigaction() failed.\n");
		die(msgbuf, 1);
	}
	window_resized = 0;


	/* create socket */
	pctx.sockfd = socket(PF_INET, SOCK_DGRAM, 0);
	if (pctx.sockfd == -1) {
		snprintf(msgbuf, MBS, "socket(): %s\n", strerror(errno));
		die(msgbuf, 1);
	}
	/* name socket (bind it to local_port) */
	local_addr.sin_family      = AF_INET;
	local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	local_addr.sin_port        = htons(local_port);
	ret = bind(pctx.sockfd, (struct sockaddr *)&local_addr, sizeof(local_addr));
	if (ret == -1) {
		snprintf(msgbuf, MBS, "bind(): %s\n", strerror(errno));
		die(msgbuf, 1);

	}
	/* get stdin fd */
	pctx.stdinfd = fileno(stdin);
	if (pctx.stdinfd == -1) {
		snprintf(msgbuf, MBS, "fileno(): %s\n", strerror(errno));
		die(msgbuf, 1);
	}

	/* file descriptor for socket is at pfds[0] */
	pfds[0].fd     = pctx.sockfd;
	pfds[0].events = POLLIN;
	nfds           = 1;
	/* file descriptor for stdin  is at pfds[1] */
	pfds[1].fd     = pctx.stdinfd;
	pfds[1].events = POLLIN;
	nfds           = 2;

	/* init alsa audio capture */
	ret = audio_init(&ad_capture, audiodevice, SND_PCM_STREAM_CAPTURE,
		             rate, channels, pfds+nfds, verbose >= 2 ? 1 : 0, fplog);
	if (ret == -1)
		die(NULL, 1);
	nfds += ad_capture.nfds;

	/* init alsa audio playback */
	ret = audio_init(&ad_playback, audiodevice, SND_PCM_STREAM_PLAYBACK,
		             rate, channels, pfds+nfds, verbose >= 2 ? 1 : 0, fplog);
	if (ret == -1)
		die(NULL, 1);
	/* we won't poll this device */
	//nfds += ad_playback.nfds;


	/* create encoder state */
	snprintf(msgbuf, MBS, "libopus: %s\n", opus_get_version_string());
	log_msg(msgbuf);
	enc = opus_encoder_create(rate, channels, OPUS_APPLICATION_VOIP, &error);
	check_opus_error_or_die(error);
#if 0
	/* set maximum bandwidth */
	bandwidth = OPUS_BANDWIDTH_WIDEBAND;
	error = opus_encoder_ctl(enc, OPUS_SET_MAX_BANDWIDTH(bandwidth));
	check_opus_error_or_die(error);
#endif
	/* set bitrate */
	error = opus_encoder_ctl(enc, OPUS_SET_BITRATE(bitrate));
	check_opus_error_or_die(error);
	/* set signal type */
	error = opus_encoder_ctl(enc, OPUS_SET_SIGNAL(OPUS_SIGNAL_VOICE));
	check_opus_error_or_die(error);


	/* welcome message */
	snprintf(msgbuf, MBS, "Welcome! Enter /h for help.\n");
	print_msg(msgbuf);
	snprintf(msgbuf, MBS, "Listening for incoming calls on udp port %d...\n", local_port);
	print_msg(msgbuf);

	/* install signal handler */
	act.sa_handler = signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	ret = sigaction(SIGINT, &act, NULL);
	if (ret == -1) {
		snprintf(msgbuf, MBS, "Error: sigaction() failed.\n");
		die(msgbuf, 1);
	}
	keep_going = 1;

	/* start audio capture */
	ret = audio_start(&ad_capture);
	if (ret == -1)
		die(NULL, 1);
	keep_going_capture = 1;

	/* reset counters */
	ic_total             = 0; /* total capture frame counter */
	ic_talk              = 0; /*  talk capture frame counter */
	ic_mute              = 0; /*  mute capture frame counter */
	total_bytes_out      = 0;
	total_bytes_in       = 0;
	peak_percent         = 0.0f;
	overruns             = 0;
	underruns            = 0;
	micgain              = 1.0f;
	micmute              = 0;

	/* main loop */
	while (keep_going || keep_going_capture) {
		/* poll: is audio_from_capture or udp_socket ready? */
		pollret = poll(pfds, nfds, -1);

		/* poll can fail and return -1 when a signal arrives. errno will be EINTR */
		if (pollret == -1) {
			if (errno == EINTR) {
				continue;
			} else {
				snprintf(msgbuf, MBS, "poll(): %s\n", strerror(errno));
				die(msgbuf, 1);
			}
		}
#if 0
		/* timeout should never happen really... */
		if (pollret == 0) {
			snprintf(msgbuf, MBS, "poll timeout\n");
			log_msg(msgbuf);
			continue;
		}
#endif
		/* window resize */
		if (window_resized) {
			endwin();
			refresh();

			getmaxyx(stdscr, nctx.rows, nctx.cols);
			clear();
			refresh();

			wresize(nctx.win, nctx.rows - 4, nctx.cols);
			wclear(nctx.win);
			snprintf(msgbuf, MBS, "Window resized to %dx%d\n", nctx.cols, nctx.rows);
			print_msg(msgbuf);

			clear_input_line();
			print_bars(VERSION);
			print_prompt(pctx.nick);

			window_resized = 0;
		}

		/** UDP SOCKET **/
		if (pfds[0].revents & POLLERR) {
			snprintf(msgbuf, MBS, "Socket POLLERR\n");
			die(msgbuf, 1);
		}

		/* process data from socket if ready */
		if (pfds[0].revents & POLLIN) {
			/* read udp_packet */
			receive_udp_packet(pctx.sockfd, &udp_packet);
			total_bytes_in += udp_packet.len;

			ret = process_packet(&pctx, &udp_packet);
			continue;
		}


		/** STDIN **/
		if (pfds[1].revents & POLLIN) {
			int len, n, ch;

			/* read an entire line, skipping special keys */
			ch = getch();
			if (ch == KEY_RESIZE ||
			    ch == KEY_DOWN || ch == KEY_UP || ch == KEY_LEFT || ch == KEY_RIGHT) {
				continue;
			}
			if (ch == 27) { /* Esc */
				keep_going = 0;
				continue;
			}
			if (ch == KEY_F(5)) {
				micmute = (micmute == 0) ? 1 : 0;
				snprintf(msgbuf, MBS, "mute: %s\n", micmute ? "on" : "off");
				print_msg(msgbuf);
				continue;
			}
#if 0
			if (ch == KEY_F(6) || ch == KEY_F(7)) {
				if (pctx.need_answer) {
					pctx.answer = (ch == KEY_F(6)) ? ANS_YES : ANS_NO;
					process_packet(&pctx, &pctx.saved_packet);
				}
				continue;
			}
#endif
			if (ch == KEY_BACKSPACE) {
				int y, x;

				if (nctx.linepos) {
					nctx.nchars--;
					nctx.linepos--;
					/* if we have a continuation byte, erase another char */
					if (nctx.linepos && (nctx.linebuf[nctx.linepos] & 0x80))
						nctx.linepos--;
					getyx(stdscr, y, x);
					move(y, x-1);
					addch(' ');
					move(y, x-1);
					refresh();
				}
				continue;
			}

			/* limit max characters read */
			if (ch != '\n' && nctx.nchars >= (nctx.cols - strlen(pctx.nick) - 3)) {
				//clear_input_line();
				//print_prompt(pctx.nick);
				pctx.tone.tone_play = (pctx.tone.tone_play > 10*ONE_OPUS_FRAME) ? pctx.tone.tone_play : 10*ONE_OPUS_FRAME;
				continue;
			}

			/* insert char */
			if (ch < 128) {
				nctx.nchars++;
				nctx.linebuf[nctx.linepos++] = ch;
				addch(ch);
				refresh();
				if (ch != '\n')
					continue;
			} else if ((ch & 0xe0) == 0xc0) { /* UTF-8 char encoded in 2 bytes */
//				int y, x;

				nctx.nchars++;
				nctx.linebuf[nctx.linepos++] = ch;
				addch(ch);
				ch = getch();
				nctx.linebuf[nctx.linepos++] = ch;
				addch(ch);
//				getyx(stdscr, y, x);
//				move(y, x-1);  /* we called addch twice, go back once */

				refresh();
				continue;
			}

			/* insert '\0' */
			nctx.linebuf[nctx.linepos++] = '\0';
			clear_input_line();
			print_prompt(pctx.nick);

			len = strlen(nctx.linebuf);

			/* chat */
			if (len > 1 && nctx.linebuf[0] != '/') {
				packet_chat(&udp_packet, nctx.linebuf, len+1);
				send_packet_to_all(pctx.nodes, pctx.sockfd, &udp_packet, &total_bytes_out);

				snprintf(msgbuf, MBS, "%s> %s", pctx.nick, nctx.linebuf);
				wattron(nctx.win, A_BOLD);
				print_msg(msgbuf);
				wattroff(nctx.win, A_BOLD);
			}

			/* command */
			if (len > 2 && nctx.linebuf[0] == '/') {
				switch (nctx.linebuf[1]) {
				case 'h':
					snprintf(msgbuf, MBS,
"help:\n"
"  /h             help\n"
"  /i             print info\n"
"  /m             toggle mic mute        [F5]\n"
"  /g gain        set software mic gain\n"
"  /b bitrate     set bitrate\n"
"  /k n           kill node n\n"
"  /c host[:port] call host\n"
"  /q             quit                   [ESC]\n"
);
					print_msg(msgbuf);
					break;
				case 'i':
					snprintf(msgbuf, MBS, "info:\n");
					print_msg(msgbuf);

					snprintf(msgbuf, MBS, "  capture:   rb[%3.0f%%]   overrun[%2d]  gain[%.2f]  mute[%s]\n",
					         100.0f * ad_capture.avail/(float)ad_capture.buffer_size,
					         overruns, micgain, micmute ? " on" : "off");
					print_msg(msgbuf);
					snprintf(msgbuf, MBS, "  playback:  rb[%3.0f%%]  underrun[%2d]\n",
					         100.0f * ad_playback.avail/(float)ad_playback.buffer_size, underruns);
					print_msg(msgbuf);

					/* get bitrate */
					error = opus_encoder_ctl(enc, OPUS_GET_BITRATE(&bitrate));
					check_opus_error_or_die(error);
					/* get bandwidth */
					error = opus_encoder_ctl(enc, OPUS_GET_BANDWIDTH(&bandwidth));
					check_opus_error_or_die(error);
					snprintf(msgbuf, MBS, "  stream:    bitrate[%5d bit/s]  bandwidth[%d Hz]\n",
					        bitrate, convert_opus_bw(bandwidth));
					print_msg(msgbuf);

					snprintf(msgbuf, MBS, "  nodes: %d\n", node_get_count(pctx.nodes));
					print_msg(msgbuf);
					for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
						if (pctx.nodes[i]) {
							snprintf(msgbuf, MBS, "    %d: %s:%-5hu %12s  pl[%3d/%4d]  pcm[len:%4d full:%d]\n",
								    i, inet_ntoa(pctx.nodes[i]->addr.sin_addr),
								    ntohs(pctx.nodes[i]->addr.sin_port), pctx.nodes[i]->nick,
								    pctx.nodes[i]->packets_lost, pctx.nodes[i]->packets_count,
								    pctx.nodes[i]->pcmlen, pctx.nodes[i]->pcmfull);
							print_msg(msgbuf);
						}
					}
					break;
				case 'm':
					micmute = (micmute == 0) ? 1 : 0;
					snprintf(msgbuf, MBS, "mute: %s\n", micmute ? "on" : "off");
					print_msg(msgbuf);
					break;
				case 'g':
					if (len > 4) {
						float gain = 1.0f;
						ret = sscanf(nctx.linebuf+3, "%f", &gain);
						if (ret == 1 && 0.0f < gain && gain <= 10.0f) {
							micgain = gain;
							snprintf(msgbuf, MBS, "mic gain: %.2f\n", micgain);
							print_msg(msgbuf);
						}
					}
					break;
				case 'b':
					if (len > 4) {
						ret = sscanf(nctx.linebuf+3, "%d", &bitrate);
						if (bitrate < 6000 || bitrate > 512000)
							bitrate = 16000;
						/* set bitrate */
						error = opus_encoder_ctl(enc, OPUS_SET_BITRATE(bitrate));
						check_opus_error_or_die(error);
						/* get bitrate */
						error = opus_encoder_ctl(enc, OPUS_GET_BITRATE(&bitrate));
						check_opus_error_or_die(error);
						snprintf(msgbuf, MBS, "bitrate: %d bit/s\n", bitrate);
						print_msg(msgbuf);
					}
					break;
				case 'k':
					if (len > 4) {
						n = atoi(nctx.linebuf+3);
						if (0 <= n && n < MAX_NUMBER_OF_NODES) {
							if (pctx.nodes[n])
								node_delete(pctx.nodes, n, 0, &pctx.tone);
						}
					}
					break;
				case 'c':
					if (len > 4) {
						char *pc;
						struct sockaddr_in addr;
						int idx, port = 8110;

						/* delete trailing '\n' */
						nctx.linebuf[len-1] = '\0';

						/* extract port */
						pc = strchr(nctx.linebuf+3, ':');
						if (pc) {
							*pc = '\0';
							pc++;
							if (*pc)
								port = atoi(pc);
						}

						/* extract address */
						pc = nctx.linebuf+3;
						while (*pc == ' ')
							pc++;

						snprintf(msgbuf, MBS, "Resolving address...\n");
						print_msg(msgbuf);
						ret = resolve_address(&addr, pc, port);
						if (ret != -1) {
							idx = node_search(pctx.nodes, addr);
							if (idx == -1)
								idx = node_add(pctx.nodes, addr, NULL, &pctx.tone);
							if (idx != -1) {
								udp_packet.addr = addr;
								packet_call(&udp_packet, pctx.nick);
								send_udp_packet(pctx.sockfd, &udp_packet);
							}
						}
					}
					break;
				case 'q':
					keep_going = 0;
					break;
				default:
					snprintf(msgbuf, MBS, "Unknown command\n");
					print_msg(msgbuf);
					break;
				}
			}
		}


		/** AUDIO CAPTURE **/
		audio_poll_descriptors_revents(&ad_capture);

		if (ad_capture.revents & POLLERR) {
			if (!(ad_capture.revents & POLLIN)) {
				snprintf(msgbuf, MBS, "Audio Capture POLLERR\n");
				die(msgbuf, 1);
			}
		}

		/* process captured audio if ready */
		if (ad_capture.revents & POLLIN && keep_going_capture) {
			/* read ad_capture.period_size frames of audio */
			ret = audio_read(&ad_capture, 1);
			if (ret < 0) {
				/* overrun, let's handle it and try again */
				overruns++;
				ret = audio_recover(&ad_capture, ret, 1);
				if (ret == 0) {
					ret = audio_start(&ad_capture);
					if (ret == -1)
						die(NULL, 1);
					ret = audio_read(&ad_capture, 0);
					if (ret < 0)
						die(NULL, 1);
				} else {
					snprintf(msgbuf, MBS, "audio_recover() on capture failed\n");
					die(msgbuf, 1);
				}
			}

			/* make sure we have read an entire frame */
			if (ad_capture.frames != ad_capture.alsabuffersize) {
				snprintf(msgbuf, MBS, "Error, we only read %ld frames instead of %lu.\n",
					    ad_capture.frames, ad_capture.alsabuffersize);
				die(msgbuf, 1);
			}

			if (!micmute) {
				/* amplify if necessary */
				if (micgain != 1.0f) {
					int32_t vl, vr;
					for (i = 0; i < ad_capture.channels * ad_capture.frames; i+=2) {
						vl = lrintf((float)ad_capture.alsabuffer[i  ] * micgain);
						vr = lrintf((float)ad_capture.alsabuffer[i+1] * micgain);
						vl = (vl < INT16_MIN) ? INT16_MIN : ((vl > INT16_MAX) ? INT16_MAX : vl);
						vr = (vr < INT16_MIN) ? INT16_MIN : ((vr > INT16_MAX) ? INT16_MAX : vr);
						ad_capture.alsabuffer[i  ] = vl;
						ad_capture.alsabuffer[i+1] = vr;
					}
				}

				/* find peak in captured audio */
				audio_find_peak(&ad_capture);
				if (peak_percent < ad_capture.peak_percent)
					peak_percent = ad_capture.peak_percent;
			}

			/* increment timeout counter, delete node after 5s without receiving packets */
			for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
				if (pctx.nodes[i] /*&& pctx.nodes[i]->knows_about_me*/) {
					pctx.nodes[i]->timeout--;
					if (pctx.nodes[i]->timeout == 0)
						node_delete(pctx.nodes, i, 1, &pctx.tone);
				}
			}

			/* call packets or table packets might have gotten lost, so every
			   2s send out call requests to those who don't know about me,
			   using relay to make sure packets get delivered */
			if ((1+ic_talk+ic_mute) % 100 == 0)
				call_new_nodes(pctx.nodes, pctx.sockfd, pctx.nick, 1);


			/* if there are no nodes connected, there is nothing to do */
			if (node_get_count(pctx.nodes) == 0) {
				ic_talk = 0;
				ic_mute = 0;
			} else {
				if (micmute) {
					/* forge udp nop packet */
					packet_nop(&udp_packet);

					/* increment mute frame counter */
					ic_mute++;
				} else {
					/* encoding */
					opus_enc_packetlen = opus_encode(enc, ad_capture.alsabuffer, ad_capture.frames, opus_enc_packet, 4000);
					if (opus_enc_packetlen < 0) {
						snprintf(msgbuf, MBS, "opus_encode() failed: %s\n", opus_strerror(opus_enc_packetlen));
						die(msgbuf, 1);
					}

					/* forge udp audio packet */
					packet_audio(&udp_packet, ic_talk, opus_enc_packet, opus_enc_packetlen);

					/* increment talk frame counter */
					ic_talk++;
				}

				/* send udp packet */
				send_packet_to_all(pctx.nodes, pctx.sockfd, &udp_packet, &total_bytes_out);
			}

			/* time to exit? */
			if (keep_going == 0) {
				keep_going_capture = 0;

				packet_end(&udp_packet);
				send_packet_to_all(pctx.nodes, pctx.sockfd, &udp_packet, &total_bytes_out);
			}

			/* mix audio */
			if (mix_audio(&pctx)) {
				/* playback decoded data */
				ret = audio_write(&ad_playback, pctx.pcm, pctx.pcmlen, 1);
				if (ret < 0) {
					/* underrun, let's handle it and try again */
					underruns++;
					ret = audio_recover(&ad_playback, ret, 1);
					if (ret == 0) {
						ret = audio_write(&ad_playback, pctx.pcm, pctx.pcmlen, 0);
						if (ret < 0)
							die(NULL, 1);
					} else {
						snprintf(msgbuf, MBS, "audio_recover() on playback failed\n");
						die(msgbuf, 1);
					}
				}
			}

			/* increment total frame counter */
			ic_total++;

			/* status bar */
			if (ic_total % 50 == 0) {
				int y, x, peakcolor;

				/* read available frames */
				ret = audio_avail(&ad_capture, 1);
				if (ret < 0)
					ad_capture.avail = ad_capture.buffer_size;
				ret = audio_avail(&ad_playback, 1);
				if (ret < 0)
					ad_playback.avail = 0;

				update_node_stats(&pctx);
				getyx(stdscr, y, x);

				if (micmute)
					peakcolor = 4;
				else
					peakcolor = (peak_percent <= 50.0f) ? 2 : ((peak_percent <= 80.0f) ? 3 : 4);

				move(nctx.rows - 2, 0);
				attron(COLOR_PAIR(1));
				printw("\r%02d:%02d  peak[", (ic_talk+ic_mute)/3000, ((ic_talk+ic_mute)/50)%60);
				attron(COLOR_PAIR(peakcolor));
				if (micmute)
					printw(" MUTE ");
				else
					printw("%5.1f%%", peak_percent);
				attron(COLOR_PAIR(1));
				printw("]  speed[o:%4.1f i:%4.1f kB/s]  pl[%d/%4.2f%%]  xrun[o:%d u:%d]",
					    total_bytes_out/1024.0, total_bytes_in/1024.0,
					    pctx.packets_lost_g,
				        pctx.packets_count_g ? 100.0f*pctx.packets_lost_g/(float)pctx.packets_count_g : 0.0f,
				        overruns, underruns);
				attroff(COLOR_PAIR(1));
				chgat(-1, A_NORMAL, 1, NULL);
				move(y, x);
				refresh();

				peak_percent    = 0.0f;
				total_bytes_out = 0;
				total_bytes_in  = 0;
			}
		}
	}

	/* check if we still have stuff on the socket */
	while ((pollret = poll(pfds, 1, 250)) > 0) {
		if (pfds[0].revents & POLLIN)
			receive_udp_packet(pctx.sockfd, &udp_packet);
	}

	/* clean up */
	ret = close(pctx.sockfd);
	ret = audio_stop(&ad_capture);
	ret = audio_close(&ad_capture);
	ret = audio_stop(&ad_playback);
	ret = audio_close(&ad_playback);
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx.nodes[i])
			node_delete(pctx.nodes, i, 0, &pctx.tone);
	}

	/* destroy encoder state */
	opus_encoder_destroy(enc);

	/* exit */
	endwin();
	return 0;
}
