/*
 * Copyright (C) 2012 Giorgio Vazzana
 *
 * This file is part of parole-conference.
 *
 * parole-conference is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * parole-conference is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PAROLE_CONFERENCE_H
#define PAROLE_CONFERENCE_H

#include <stdint.h>
#include <stddef.h>
#include <netinet/in.h>
#include <opus.h>

extern const unsigned int rate, channels;

/* A CALL udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 * Payload:
 *   nick            (NICKLEN bytes)
 */
#define UDP_PACKET_TYPE_CALL    0x0000

/* A REFUSE udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 */
#define UDP_PACKET_TYPE_REFUSE  0x0001

/* A TABLE udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   table_size      (uint16, be)
 * Payload:
 *   my_nick         (NICKLEN bytes)
 *  repeat table_size times for all other nodes except caller:
 *   nick            (NICKLEN bytes)
 *   addr            (uint32, be)
 *   port            (uint16, be)
 */
#define UDP_PACKET_TYPE_TABLE   0x0002

/* An AUDIO udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   sequence_number (uint32, be)
 * Payload:
 *   data            (4000 byte max)
 *
 * The field sequence_number must start at 0 for the first packet and increase
 * monotonically. The payload data is the opus packet coming out of the encoder.
 */
#define UDP_PACKET_TYPE_AUDIO   0x0003

/* A CHAT udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 * Payload:
 *   data            (variable size)
 */
#define UDP_PACKET_TYPE_CHAT    0x0004

/* A NOP udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 */
#define UDP_PACKET_TYPE_NOP     0x0005

/* A RELAY udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   dest_addr       (uint32, be)
 *   dest_port       (uint16, be)
 * Payload:
 *   packet to relay (variable size)
 */
#define UDP_PACKET_TYPE_RELAY   0x0006

/* A RELAYED udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   source_addr     (uint32, be)
 *   source_port     (uint16, be)
 * Payload:
 *   packet relayed (variable size)
 */
#define UDP_PACKET_TYPE_RELAYED 0x0007

/* An END udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 */
#define UDP_PACKET_TYPE_END     0x0008

#define UDP_PACKET_AUDIO_HEADER_LEN  6
#define UDP_PACKET_AUDIO_PAYLOAD_LEN (4000)
#define UDP_PACKET_MAXLEN            (UDP_PACKET_AUDIO_HEADER_LEN + UDP_PACKET_AUDIO_PAYLOAD_LEN)

struct udp_packet_st {
	struct sockaddr_in   addr;
	uint8_t              data[UDP_PACKET_MAXLEN];
	size_t               len;
};

/* maximum length of nickname */
#define NICKLEN             16

/* node timeout (in units of 20ms): if we don't receive packets from a node
 * for more than TIMEOUT/50 seconds, then that node will be deleted.
 */
#define TIMEOUT             300

/* one opus audio frame is 20ms at 48kHz. Note that 960 are alsa frames,
 * and need to be multiplied by channels to get the number of samples.
 */
#define ONE_OPUS_FRAME      960

/* space for 15 opus frames. These are alsa frames, and need to be
 * multiplied by channels to get the number of samples.
 */
#define PCMBUFLEN           (15*ONE_OPUS_FRAME)

struct node {
	struct sockaddr_in   addr;
	char                 nick[NICKLEN];

	int                  knows_about_me;
	int                  timeout;
	uint32_t             sequence_number;
	uint32_t             last_sequence_number;

	/* opus decoder */
	OpusDecoder         *dec;
	opus_int16           pcm[2*PCMBUFLEN];
	opus_int32           pcmlen; /* in frames */
	int                  pcmfull;
	int                  packets_count;
	int                  packets_lost;
};

/* maximum number of nodes (excluding our node) of the network */
#define MAX_NUMBER_OF_NODES 9

/* tone buffer, 1s, in alsa frames */
#define TONEBUFLEN          (50*ONE_OPUS_FRAME)

struct tone {
	opus_int16           pcm[2*TONEBUFLEN];
	int                  tone_pos;  /* in frames */
	int                  tone_play; /* in frames */
};

struct PContext {
	/* my nickname */
	char                 nick[NICKLEN];

	/* file descriptors */
	int                  sockfd;
	int                  stdinfd;

#if 0
	/* options */
	int                  autoaccept;

	/* answer to a question */
	struct udp_packet_st saved_packet;
	int                  need_answer;
#define ANS_NOTPRESENT 0
#define ANS_YES        1
#define ANS_NO         2
	int                  answer;
#endif

	/* array of nodes */
	struct node         *nodes[MAX_NUMBER_OF_NODES];

	/* tone */
	struct tone          tone;

	/* decoded pcm data */
	opus_int16           pcm[2*ONE_OPUS_FRAME];
	opus_int32           pcmlen; /* in frames */

	/* stats */
	int                  packets_lost_g;
	int                  packets_count_g;
};

#endif
